#include <cstdlib>
#include <fstream>
#include <stdlib.h>
#include <iostream>
#include "TPAstructs.h"
#include "TPAfuncs.h"
#include <stdio.h>
#include <string>
#include <sstream>
#include <windows.h>

using namespace std;

int main(int argc, char *argv[]) {

	if(argv[1]==NULL) {
	 std::cout<<"This is a drag n' drop utility! A beta (v130 or v140) TPA"<<std::endl;
	 std::cout<<" file must be dropped onto window. Press [enter key] to exit.";
	 std::cin.get();
	 return 0;
	}

	std::string oldFileName;	// the file name program registered
	std::string TPAfilename;	// the file name portion
	std::string TPAextension;	// the extension portion
	std::string new150tpa;  // will be name of new 150 TPA file
	std::string temp150tpa;  // will be name of temp 150 TPA file
	std::string oldFileNameBackup;  // will be name of temp 150 TPA file

	oldFileName=argv[1];
	TPAextension = oldFileName;
	TPAfilename = oldFileName;

	TPAextension = TPAextension.erase(0, TPAextension.length() - 4);
	TPAfilename = TPAfilename.erase(TPAfilename.length() - 4, 4);


	if(TPAextension != ".tpa") {
	 std::cout<<"Incorrect file_type dropped onto window"<<std::endl;
	 std::cout<<"Press [enter key] to exit.";
	 std::cin.get();
	 return 0;
	}

	new150tpa = TPAfilename + ".new";
	temp150tpa = TPAfilename + ".tmp";
	oldFileNameBackup = oldFileName +".bak";

	FILE *oldTPA;
	FILE *newTPA;
	FILE *tempTPA;

	oldTPA = fopen(oldFileName.c_str(), "rb");
	unsigned long oldSampletableOffset;


// First, we take care of the main TPA header

    std::cout<<"Reading main TPA Header..."<<std::endl;

	unsigned long oldTPAversion;
	unsigned long oldNumSamples;
	unsigned long oldIDtableOffset;
	unsigned long oldNumCollisions;
	unsigned long oldFoleyOffset;
	unsigned long oldNumSoundIDs;

	TPAheader oldTPAheader;

	readTPAheader( oldTPAheader, oldTPA);

	oldTPAversion = oldTPAheader.TPAversion;
	oldNumSamples = oldTPAheader.numSoundSamples;
	oldIDtableOffset = oldTPAheader.IDtableOffset; // needs to be updated later!
	oldNumCollisions = oldTPAheader.numCollisions;
	oldFoleyOffset = oldTPAheader.FoleyTableOffset; // needs to be updated later!
	oldNumSoundIDs = oldTPAheader.numSoundIDs;

	if (oldTPAversion == 0x150) {
	 std::cout<<"File is already v150! No need to update it!"<<std::endl;
	 std::cout<<"Press [enter key] to exit.";
	 std::cin.get();
	 return 0;
	}

	if (oldTPAversion < 0x130) {
	 std::cout<<"File version is earlier than v130! Can't be processed!"<<std::endl;
	 std::cout<<"Press [enter key] to exit.";
	 std::cin.get();
	 return 0;
	}

	oldTPAheader.TPAversion = 0x150;

	tempTPA=fopen(temp150tpa.c_str(), "wb");


// Second, we go through the Sample (CAU) Table

    std::cout<<"Updating Sound Samples (CAU files)..."<<std::endl;

	unsigned long tempOffsets[oldNumSamples+1][3];
	tempOffsets[0][1] = 0;

// The offset to the Sample Table is the offset to the entry in the Sample Table for the first Sample ID

	fseek(oldTPA, oldIDtableOffset+4, SEEK_SET);
	unsigned long oldSampleTableOffset;
	fread(&oldSampleTableOffset, sizeof(unsigned long), 1, oldTPA);
	fseek(oldTPA, oldSampleTableOffset, SEEK_SET);


	TPAsampleHeader oldTPAsampleHeader;
	unsigned long   captionOffset; // only in v110 files, 0x00 if no captions, 0x24 if captions

	char sampleBuffer;
	char captionBuffer;
	unsigned long captionBlockLength;
    
	for (int i = 0; i < oldNumSamples; i++) {

	tempOffsets[i][0] = ftell(oldTPA);

		readTPAsampleHeader( oldTPAsampleHeader, oldTPA);
		if (oldTPAsampleHeader.CAUversion == 100) {
			captionOffset = 0;
			oldTPAsampleHeader.CAUversion = 110;
			oldTPAsampleHeader.SampleDataOffset += 4;
		}
		else {
			fread(&captionOffset, sizeof(unsigned long), 1, oldTPA);
		}
		writeTPAsampleHeader( oldTPAsampleHeader, tempTPA);
		fwrite(&captionOffset, sizeof(unsigned long), 1, tempTPA);
		if (captionOffset != 0) {
			fread(&captionBlockLength, sizeof(unsigned long), 1, oldTPA);
			fwrite(&captionBlockLength, sizeof(unsigned long), 1, tempTPA);
			for (int j = 0; j < captionBlockLength; j++) {
				fread(&captionBuffer, sizeof(char), 1, oldTPA);
				fwrite(&captionBuffer, sizeof(char), 1, tempTPA);
			}
		}
		for (int j = 0; j < oldTPAsampleHeader.SampleDataLength; j++) {
			fread(&sampleBuffer, sizeof(char), 1, oldTPA);
			fwrite(&sampleBuffer, sizeof(char), 1, tempTPA);
		}

	tempOffsets[i+1][1] = ftell(tempTPA);
	tempOffsets[i][2] = tempOffsets[i+1][1] - tempOffsets[i][1];

	}


// Let's not forget now to update the Foley Table Offset in the main header by now!
// We should be at that address right now

	oldTPAheader.FoleyTableOffset = tempOffsets[oldNumSamples][1] + 24 + 56 * oldNumSoundIDs;

// Now, if needed, we go through the Foley Table

	if (oldNumCollisions > 0) {

        std::cout<<"Updating Foley Table..."<<std::endl;

		FoleyTableEntry oldFoleyTableEntry;
		transferFunctionTable oldTransferFunctionTable;
		float	minVelocity;

		for (int i = 0; i < oldNumCollisions; i++) {
			readFoleyTableEntry( oldFoleyTableEntry, oldTPA);
			writeFoleyTableEntry( oldFoleyTableEntry, tempTPA);
			for (int j = 0; j < 3; j++) {
				readTransferFunctionTable(  oldTransferFunctionTable, oldTPA);
				writeTransferFunctionTable(  oldTransferFunctionTable, tempTPA);
				if (oldTPAversion == 0x130) {
					minVelocity = 0.0;
				}
				else {
					fread(&minVelocity, sizeof(float), 1, oldTPA);
				}
				fwrite(&minVelocity, sizeof(float), 1, tempTPA);
			}
		}
	}

	fclose(tempTPA);

// Now we can start writing to our new TPA file;

    std::cout<<"Writing updated TPA Header..."<<std::endl;

	newTPA=fopen(new150tpa.c_str(), "wb");
	tempTPA=fopen(temp150tpa.c_str(), "rb");

	writeTPAheader( oldTPAheader, newTPA);


// At this point, we can go back and go through the Sound ID Table

    std::cout<<"Updating Sound IDs Table..."<<std::endl;

	fseek(oldTPA, oldIDtableOffset, SEEK_SET);


	float	SampleDuration; // in seconds; Absent in v150 files
	float	Attenuation; // in dB; Absent in v130 files
	float	MasterVolume; // in dB; Absent in v130 files

	TPAidHeader oldTPAidHeader;

	unsigned long oldSoundID;
	unsigned long oldSampleOffset;
	unsigned long oldSampleLength;
	unsigned long	newSampleOffset; // we'll read it from the temp file

	for (int i = 0; i < oldNumSoundIDs; i++) {
		readTPAidHeader( oldTPAidHeader, oldTPA);
		oldSoundID = oldTPAidHeader.SoundID;
		oldSampleOffset = oldTPAidHeader.SampleOffset; // needs to be updated later!
		oldSampleLength = oldTPAidHeader.SampleLength; // may need to be updated too!

// Now here comes the tricky part!!!
// We need to use the oldSampleOffset to find the corresponding Sample Offset and Sample Length in our temp file
// and calculate the new Sample Offset for our new TPA file!

		for (int j = 0; j < oldNumSamples; j++) {
			if (tempOffsets[j][0] == oldSampleOffset) {
				newSampleOffset = tempOffsets[j][1];
				oldTPAidHeader.SampleOffset = newSampleOffset + 24 + 56 * oldNumSoundIDs;
				oldTPAidHeader.SampleLength = tempOffsets[j][2];
			}
		}

		fread(&SampleDuration, sizeof(float), 1, oldTPA);
		if (oldTPAversion == 0x130) {
			Attenuation = 0.0;
			MasterVolume = 0.0;
		}
		else {
			fread(&Attenuation, sizeof(float), 1, oldTPA);
			fread(&MasterVolume, sizeof(float), 1, oldTPA);
		}
		writeTPAidHeader( oldTPAidHeader, newTPA);
		fwrite(&Attenuation, sizeof(float), 1, newTPA);
		fwrite(&MasterVolume, sizeof(float), 1, newTPA);

// We now copy the corresponding CAU header (or TPAsampleHeader), which has been stored in our temp TPA file!

		fseek(tempTPA, newSampleOffset, SEEK_SET);

		readTPAsampleHeader( oldTPAsampleHeader, tempTPA);

		if (oldTPAsampleHeader.CAUversion == 100) {
			captionOffset = 0;
			oldTPAsampleHeader.CAUversion = 110;
			oldTPAsampleHeader.SampleDataOffset += 4;
		}
		else {
			fread(&captionOffset, sizeof(unsigned long), 1, tempTPA);
		}

		writeTPAsampleHeader( oldTPAsampleHeader, newTPA);
		fwrite(&captionOffset, sizeof(unsigned long), 1, newTPA);

	}

	fclose(oldTPA);

// And finally, we append the contents of the temp TPA file to our new TPA file


    std::cout<<"Appending Sound Samples (CAU) Table and Foley Table..."<<std::endl;

	unsigned long appendBytes;
	char appendBuffer;

	fseek(tempTPA, 0, SEEK_END);
	appendBytes = ftell(tempTPA);
	fseek(tempTPA, 0, SEEK_SET);

	for (int i = 0; i < appendBytes; i++) {
	      fread(&appendBuffer, sizeof(char), 1, tempTPA);
	      fwrite(&appendBuffer, sizeof(char), 1, newTPA);
	}
	fclose(tempTPA);
	fclose(newTPA);

	DeleteFile(temp150tpa.c_str());
	int renOldTPA = rename(oldFileName.c_str(), oldFileNameBackup.c_str());
	int renNewTPA = rename(new150tpa.c_str(), oldFileName.c_str());

    std::cout<<"Done! TPA file has been updated to latest version. Press [enter key] to exit.";
    std::cin.get();
    return 0;
}
